#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <math.h>
#include <GL/glut.h>
#include "graphics.h"

static int winID;
static int xsize, ysize;

static GLboolean inStereo;
static unsigned char *snapshBuf;

/* private function prototypes */
static void write_ppmfile( char *fname, unsigned char *pixels, int w, int h );

GLfloat mat_specular[] = { 0.5, 0.5, 0.5, 1.0}; /* RGBA */
GLfloat mat_shininess[] = { 50.0 }; /* specular exponent */
GLfloat light_position[] = { 1.0, 1.0, 1.0, 1.0 }; /* homogeneous coords */

/***********************************************************************
** Function:  grInit
** Arguments: c = argument count of v
**            v = string arguments to glutInit (typically from command line)
**            x = X position of window
**            y = Y position of window
**            w = width of mono window
**            h = height of viewing area of window
**            s = if TRUE, produce window suitable for stereo
**                     (implies doubling width)
** Returns:   
***********************************************************************/
void
grInit( int *c, char *v[], int x, int y, int w, int h, GLboolean s )
{
    int bytes;

    glutInitWindowSize( ( s ) ? 2 * w : w, h );
    if( s )
      glutInitWindowPosition( x, 1024 - ( y )); /* scrn y = 1024? */
    glutInit( c, v );
    glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );

    winID = glutCreateWindow( "NoTitle" );

    bytes = ( ( s ) ? ( 2 * w ) : w ) * ( h ) * 3;
    snapshBuf = (unsigned char *) malloc( bytes * sizeof( unsigned char ));
    if( snapshBuf == NULL ) {
      fprintf( stderr, "Unable to allocate space for snapshot - disabling\n" );
    }

    xsize = w;
    ysize = h;
    inStereo = s;

    glClearColor( 0.0, 0.0, 0.0, 0.0 );

    glEnable( GL_DEPTH_TEST );

    glMaterialfv( GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular );
    glMaterialfv( GL_FRONT_AND_BACK, GL_SHININESS, mat_shininess );
    glColorMaterial( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );
    glEnable(GL_COLOR_MATERIAL);
    glShadeModel(GL_SMOOTH);

    /* light is stored in eye coordinates, so it moves with you */
    glMatrixMode( GL_MODELVIEW);
    glLoadIdentity();
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glLightfv(GL_LIGHT0, GL_POSITION, light_position);

    glMatrixMode( GL_PROJECTION );
    glLoadIdentity( );
    gluPerspective( 45 * 0.75, 4.0 / 3.0, 0.01, 12.0 );

    return;
}

/***********************************************************************
** Function:  grClose
** Arguments: 
** Returns:   
***********************************************************************/
void
grClose( void )
{
    return;
}

/***********************************************************************
** Function:  grStartFrame
** Arguments: 
** Returns:   
***********************************************************************/
void
grStartFrame( void )
{
    glutSetWindow( winID );

    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    return;
}

/***********************************************************************
** Function:  grLeftView
** Arguments: left_eye_from_world = transformation matrix
** Returns:   
***********************************************************************/
void
grLeftView( GLdouble *left_eye_from_world )
{
    glViewport( 0, 0, xsize, ysize );
    glMatrixMode( GL_MODELVIEW );
    glLoadMatrixd( left_eye_from_world );

    return;
}



/***********************************************************************
** Function:  grRightView
** Arguments: 
** Returns:   
***********************************************************************/
void
grRightView( GLdouble *right_eye_from_world )
{
    glViewport( xsize, 0, xsize, ysize );
    glMatrixMode( GL_MODELVIEW );
    glLoadMatrixd( right_eye_from_world );

    return;
}


/***********************************************************************
** Function:  grEndFrame
** Arguments: 
** Returns:   
** Notes:     
***********************************************************************/
void
grEndFrame( void )
{
    glutSwapBuffers( );
}

/*
** Writes a buffer as a ppm (portable pixel map) file.  The buffer
** contains pixels in xBGR format.
*/
static void
write_ppmfile( char *fname, unsigned char *pixels, int width, int height )
{
  FILE *ppmfile;
  int i, j, c, counts[11], total;
  unsigned char *curpix;

  /* Prepare for printing stats during filesaving */
  for( i = 0; i <= 10; i++ )
    counts[i] = (int)( ( i / 10.0 ) * width * height);

  /* Create the output file */
  if(( ppmfile = fopen( fname, "w" )) == NULL ) {
    fprintf( stderr, "write_ppmfile: cannot create '%s'\n", fname );
    perror( "fopen" );
    fflush( stderr );
    return;
  }

  /* Write the ppm header */
  fprintf( ppmfile, "P6\n" );
  fprintf( ppmfile, "%d %d\n255\n", width, height );

  /*
  ** Dump the pixel buffer to file.  I don't know how to make OpenGL read
  ** out the frame buffer right-side up, so this code flips vertically.
  */
  fprintf( stderr, "Writing snapshot file '%s'\n", fname);
  fflush( stderr );
  for( i = height - 1, c = 0, total = 0; i >= 0; i-- ) {
    curpix = &pixels[ i * width * 3 ];
    for( j = 0; j < width; j++, total++ ) {
      fprintf( ppmfile, "%c", (char) ( *curpix++ ) ); /* R */
      fprintf( ppmfile, "%c", (char) ( *curpix++ ) ); /* G */
      fprintf( ppmfile, "%c", (char) ( *curpix++ ) ); /* B */
      if( total > counts[c] ) {
	fprintf( stderr, "%d%%...", c * 10 );
	fflush( stderr );
	c++;
      }
    }
  }

  /* Close file */
  fclose( ppmfile );

  fprintf( stderr, "\nDone.\n");
  fflush( stderr );

  return;
}

/* takes a snapshot picture of the display window and saves it to a file */
void
grSnapshot( char *fnameBase )
{
    static int width, height;
    static char fileName[80];

    if( snapshBuf != NULL ) {
      width = xsize;
      height = ysize;
      if( inStereo )
	width *= 2;

      sprintf( fileName, "%s.ppm", fnameBase );
      glutSetWindow( winID );
      glReadPixels( 0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, snapshBuf );
      write_ppmfile( fileName, snapshBuf, width, height );
    }

    return;
}
